import argparse
import logging
import sys

from logdetective.constants import DEFAULT_ADVISOR
from logdetective.utils import (
    process_log, initialize_model, retrieve_log_content, format_snippets, compute_certainty)
from logdetective.extractors import LLMExtractor, DrainExtractor

LOG = logging.getLogger("logdetective")


def setup_args():
    """ Setup argument parser and return arguments. """
    parser = argparse.ArgumentParser("logdetective")
    parser.add_argument("file", type=str,
                        default="", help="The URL or path to the log file to be analyzed.")
    parser.add_argument("-M", "--model",
                        help="The path or Hugging Face name of the language model for analysis.",
                        type=str, default=DEFAULT_ADVISOR)
    parser.add_argument("-F", "--filename_suffix",
                        help="Suffix of the model file name to be retrieved from Hugging Face.\
                            Makes sense only if the model is specified with Hugging Face name.",
                        default="Q4_K_S.gguf")
    parser.add_argument("-n", "--no-stream", action='store_true')
    parser.add_argument("-S", "--summarizer", type=str, default="drain",
                        help="Choose between LLM and Drain template miner as the log summarizer.\
                                LLM must be specified as path to a model, URL or local file.")
    parser.add_argument("-N", "--n_lines", type=int,
                        default=8, help="The number of lines per chunk for LLM analysis.\
                            This only makes sense when you are summarizing with LLM.")
    parser.add_argument("-C", "--n_clusters", type=int, default=8,
                        help="Number of clusters for Drain to organize log chunks into.\
                            This only makes sense when you are summarizing with Drain")
    parser.add_argument("-v", "--verbose", action='count', default=0)
    parser.add_argument("-q", "--quiet", action='store_true')
    return parser.parse_args()


def main():
    """Main execution function."""
    args = setup_args()

    if args.verbose and args.quiet:
        sys.stderr.write("Error: --quiet and --verbose is mutually exclusive.\n")
        sys.exit(2)

    # Logging facility setup
    log_level = logging.INFO
    if args.verbose >= 1:
        log_level = logging.DEBUG
    if args.quiet:
        log_level = 0

    logging.basicConfig(stream=sys.stdout)
    LOG.setLevel(log_level)

    # Primary model initialization
    try:
        model = initialize_model(args.model, filename_suffix=args.filename_suffix,
                                 verbose=args.verbose > 2)
    except ValueError as e:
        LOG.error(e)
        LOG.error("You likely do not have enough memory to load the AI model")
        sys.exit(3)

    # Log file summarizer selection and initialization
    if args.summarizer == "drain":
        extractor = DrainExtractor(args.verbose > 1, context=True, max_clusters=args.n_clusters)
    else:
        summarizer_model = initialize_model(args.summarizer, verbose=args.verbose > 2)
        extractor = LLMExtractor(summarizer_model, args.verbose > 1)

    LOG.info("Getting summary")

    try:
        log = retrieve_log_content(args.file)
    except ValueError as e:
        # file does not exists
        LOG.error(e)
        sys.exit(4)
    log_summary = extractor(log)

    ratio = len(log_summary) / len(log.split('\n'))

    LOG.info("Compression ratio: %s", ratio)

    LOG.info("Analyzing the text")

    log_summary = format_snippets(log_summary)
    LOG.info("Log summary: \n %s", log_summary)

    stream = True
    if args.no_stream:
        stream = False
    response = process_log(log_summary, model, stream)
    probs = []
    print("Explanation:")
    # We need to extract top token probability from the response
    # CreateCompletionResponse structure of llama-cpp-python.
    # `compute_certainty` function expects list of dictionaries with form
    # { 'logprob': <float> } as expected from the OpenAI API.

    if args.no_stream:
        print(response["choices"][0]["text"])
        probs = [{'logprob': e} for e in response['choices'][0]['logprobs']['token_logprobs']]

    else:
        # Stream the output
        for chunk in response:
            if isinstance(chunk["choices"][0]["logprobs"], dict):
                probs.append({'logprob': chunk["choices"][0]["logprobs"]['token_logprobs'][0]})
            delta = chunk['choices'][0]['text']
            print(delta, end='', flush=True)
    certainty = compute_certainty(probs)

    print(f"\nResponse certainty: {certainty:.2f}%\n")


if __name__ == "__main__":
    main()



import asyncio
import json
import logging
import os
from typing import List, Annotated

from llama_cpp import CreateCompletionResponse
from fastapi import FastAPI, HTTPException, Depends, Header
from fastapi.responses import StreamingResponse
import requests

from logdetective.constants import (
    PROMPT_TEMPLATE,
    SNIPPET_PROMPT_TEMPLATE,
    PROMPT_TEMPLATE_STAGED,
    SNIPPET_DELIMITER,
)
from logdetective.extractors import DrainExtractor
from logdetective.utils import validate_url, compute_certainty
from logdetective.server.models import BuildLog, Response, StagedResponse
from logdetective.server.utils import load_server_config

LOG = logging.getLogger("logdetective")

LLM_CPP_HOST = os.environ.get("LLAMA_CPP_HOST", "localhost")
LLM_CPP_SERVER_ADDRESS = f"http://{LLM_CPP_HOST}"
LLM_CPP_SERVER_PORT = os.environ.get("LLAMA_CPP_SERVER_PORT", 8000)
LLM_CPP_SERVER_TIMEOUT = os.environ.get("LLAMA_CPP_SERVER_TIMEOUT", 600)
LOG_SOURCE_REQUEST_TIMEOUT = os.environ.get("LOG_SOURCE_REQUEST_TIMEOUT", 60)
API_TOKEN = os.environ.get("LOGDETECTIVE_TOKEN", None)
SERVER_CONFIG_PATH = os.environ.get("LOGDETECTIVE_SERVER_CONF", None)
LLM_API_TOKEN = os.environ.get("LLM_API_TOKEN", None)

SERVER_CONFIG = load_server_config(SERVER_CONFIG_PATH)


def requires_token_when_set(authentication: Annotated[str | None, Header()] = None):
    """
    FastAPI Depend function that expects a header named Authentication

    If LOGDETECTIVE_TOKEN env var is set, validate the client-supplied token
    otherwise ignore it
    """
    if not API_TOKEN:
        LOG.info("LOGDETECTIVE_TOKEN env var not set, authentication disabled")
        # no token required, means local dev environment
        return
    token = None
    if authentication:
        try:
            token = authentication.split(" ", 1)[1]
        except (ValueError, IndexError):
            LOG.warning(
                "Authentication header has invalid structure (%s), it should be 'Bearer TOKEN'",
                authentication,
            )
            # eat the exception and raise 401 below
            token = None
        if token == API_TOKEN:
            return
    LOG.info(
        "LOGDETECTIVE_TOKEN env var is set (%s), clien token = %s", API_TOKEN, token
    )
    raise HTTPException(status_code=401, detail=f"Token {token} not valid.")


app = FastAPI(dependencies=[Depends(requires_token_when_set)])


def process_url(url: str) -> str:
    """Validate log URL and return log text."""
    if validate_url(url=url):
        try:
            log_request = requests.get(url, timeout=int(LOG_SOURCE_REQUEST_TIMEOUT))
        except requests.RequestException as ex:
            raise HTTPException(
                status_code=400, detail=f"We couldn't obtain the logs: {ex}"
            ) from ex

        if not log_request.ok:
            raise HTTPException(
                status_code=400,
                detail="Something went wrong while getting the logs: "
                f"[{log_request.status_code}] {log_request.text}",
            )
    else:
        LOG.error("Invalid URL received ")
        raise HTTPException(status_code=400, detail=f"Invalid log URL: {url}")

    return log_request.text


def mine_logs(log: str) -> List[str]:
    """Extract snippets from log text"""
    extractor = DrainExtractor(
        verbose=True, context=True, max_clusters=SERVER_CONFIG.extractor.max_clusters
    )

    LOG.info("Getting summary")
    log_summary = extractor(log)

    ratio = len(log_summary) / len(log.split("\n"))
    LOG.debug("Log summary: \n %s", log_summary)
    LOG.info("Compression ratio: %s", ratio)

    return log_summary


async def submit_text(
    text: str,
    max_tokens: int = -1,
    log_probs: int = 1,
    stream: bool = False,
    model: str = "default-model",
):
    """Submit prompt to LLM.
    max_tokens: number of tokens to be produces, 0 indicates run until encountering EOS
    log_probs: number of token choices to produce log probs for
    """
    LOG.info("Analyzing the text")
    data = {
        "prompt": text,
        "max_tokens": max_tokens,
        "logprobs": log_probs,
        "stream": stream,
        "model": model,
    }

    headers = {"Content-Type": "application/json"}

    if LLM_API_TOKEN:
        headers["Authorization"] = f"Bearer {LLM_API_TOKEN}"

    try:
        # Expects llama-cpp server to run on LLM_CPP_SERVER_ADDRESS:LLM_CPP_SERVER_PORT
        response = requests.post(
            f"{LLM_CPP_SERVER_ADDRESS}:{LLM_CPP_SERVER_PORT}/v1/completions",
            headers=headers,
            data=json.dumps(data),
            timeout=int(LLM_CPP_SERVER_TIMEOUT),
            stream=stream,
        )
    except requests.RequestException as ex:
        raise HTTPException(
            status_code=400, detail=f"Llama-cpp query failed: {ex}"
        ) from ex
    if not stream:
        if not response.ok:
            raise HTTPException(
                status_code=400,
                detail="Something went wrong while getting a response from the llama server: "
                f"[{response.status_code}] {response.text}",
            )
        try:
            response = json.loads(response.text)
        except UnicodeDecodeError as ex:
            LOG.error("Error encountered while parsing llama server response: %s", ex)
            raise HTTPException(
                status_code=400,
                detail=f"Couldn't parse the response.\nError: {ex}\nData: {response.text}",
            ) from ex
    else:
        return response

    return CreateCompletionResponse(response)


@app.post("/analyze", response_model=Response)
async def analyze_log(build_log: BuildLog):
    """Provide endpoint for log file submission and analysis.
    Request must be in form {"url":"<YOUR_URL_HERE>"}.
    URL must be valid for the request to be passed to the LLM server.
    Meaning that it must contain appropriate scheme, path and netloc,
    while lacking  result, params or query fields.
    """
    log_text = process_url(build_log.url)
    log_summary = mine_logs(log_text)
    response = await submit_text(PROMPT_TEMPLATE.format(log_summary))
    certainty = 0

    if "logprobs" in response["choices"][0]:
        try:
            certainty = compute_certainty(
                response["choices"][0]["logprobs"]["content"][0]["top_logprobs"]
            )
        except ValueError as ex:
            LOG.error("Error encountered while computing certainty: %s", ex)
            raise HTTPException(
                status_code=400,
                detail=f"Couldn't compute certainty with data:\n"
                f"{response["choices"][0]["logprobs"]["content"][0]["top_logprobs"]}",
            ) from ex

    return Response(explanation=response, response_certainty=certainty)


@app.post("/analyze/staged", response_model=StagedResponse)
async def analyze_log_staged(build_log: BuildLog):
    """Provide endpoint for log file submission and analysis.
    Request must be in form {"url":"<YOUR_URL_HERE>"}.
    URL must be valid for the request to be passed to the LLM server.
    Meaning that it must contain appropriate scheme, path and netloc,
    while lacking  result, params or query fields.
    """
    log_text = process_url(build_log.url)
    log_summary = mine_logs(log_text)

    # Process snippets asynchronously
    analyzed_snippets = await asyncio.gather(
        *[submit_text(SNIPPET_PROMPT_TEMPLATE.format(s)) for s in log_summary]
    )

    analyzed_snippets = [
        {"snippet": e[0], "comment": e[1]} for e in zip(log_summary, analyzed_snippets)
    ]

    final_prompt = PROMPT_TEMPLATE_STAGED.format(
        f"\n{SNIPPET_DELIMITER}\n".join(
            [
                f"[{e["snippet"]}] : [{e["comment"]["choices"][0]["text"]}]"
                for e in analyzed_snippets
            ]
        )
    )

    final_analysis = await submit_text(final_prompt)
    print(final_analysis)
    certainty = 0

    if "logprobs" in final_analysis["choices"][0]:
        try:
            certainty = compute_certainty(
                final_analysis["choices"][0]["logprobs"]["content"][0]["top_logprobs"]
            )
        except ValueError as ex:
            LOG.error("Error encountered while computing certainty: %s", ex)
            raise HTTPException(
                status_code=400,
                detail=f"Couldn't compute certainty with data:\n"
                f"{final_analysis["choices"][0]["logprobs"]["content"][0]["top_logprobs"]}",
            ) from ex

    return StagedResponse(
        explanation=final_analysis,
        snippets=analyzed_snippets,
        response_certainty=certainty,
    )


@app.post("/analyze/stream", response_class=StreamingResponse)
async def analyze_log_stream(build_log: BuildLog):
    """Stream response endpoint for Logdetective.
    Request must be in form {"url":"<YOUR_URL_HERE>"}.
    URL must be valid for the request to be passed to the LLM server.
    Meaning that it must contain appropriate scheme, path and netloc,
    while lacking  result, params or query fields.
    """
    log_text = process_url(build_log.url)
    log_summary = mine_logs(log_text)
    stream = await submit_text(PROMPT_TEMPLATE.format(log_summary), stream=True)

    return StreamingResponse(stream)



from typing import List, Dict, Optional
from pydantic import BaseModel


class BuildLog(BaseModel):
    """Model of data submitted to API."""

    url: str


class Response(BaseModel):
    """Model of data returned by Log Detective API

    explanation: CreateCompletionResponse
        https://llama-cpp-python.readthedocs.io/en/latest/api-reference/#llama_cpp.llama_types.CreateCompletionResponse
    response_certainty: float
    """

    explanation: Dict
    response_certainty: float


class StagedResponse(Response):
    """Model of data returned by Log Detective API when called when staged response
    is requested. Contains list of reponses to prompts for individual snippets.

    explanation: CreateCompletionResponse
        https://llama-cpp-python.readthedocs.io/en/latest/api-reference/#llama_cpp.llama_types.CreateCompletionResponse
    response_certainty: float
    snippets:
        list of dictionaries { 'snippet' : '<original_text>, 'comment': CreateCompletionResponse }
    """

    snippets: List[Dict[str, str | Dict]]


class InferenceConfig(BaseModel):
    """Model for inference configuration of logdetective server."""

    max_tokens: int = -1
    log_probs: int = 1

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        self.max_tokens = data.get("max_tokens", -1)
        self.log_probs = data.get("log_probs", 1)


class ExtractorConfig(BaseModel):
    """Model for extractor configuration of logdetective server."""

    context: bool = True
    max_clusters: int = 8
    verbose: bool = False

    def __init__(self, data: Optional[dict] = None):
        super().__init__()
        if data is None:
            return

        self.context = data.get("context", True)
        self.max_clusters = data.get("max_clusters", 8)
        self.verbose = data.get("verbose", False)


class Config(BaseModel):
    """Model for configuration of logdetective server."""

    inference: InferenceConfig = InferenceConfig()
    extractor: ExtractorConfig = ExtractorConfig()

    def __init__(self, data: Optional[dict] = None):
        super().__init__()

        if data is None:
            return

        self.inference = InferenceConfig(data.get("inference"))
        self.extractor = ExtractorConfig(data.get("extractor"))



import yaml
from logdetective.server.models import Config


def load_server_config(path: str | None) -> Config:
    """Load configuration file for logdetective server.
    If no path was provided, or if the file doesn't exist, return defaults.
    """
    if path is not None:
        try:
            with open(path, "r") as config_file:
                return Config(yaml.safe_load(config_file))
        except FileNotFoundError:
            pass
    return Config()
